unit main_f;

interface

uses basep4,
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ComCtrls, ImgList, ExtCtrls;

type
  TFormMain = class(TForm)
    TimerChute: TTimer; // Utilis pour l'animation de chute du pion
    TimerAlignement: TTimer; // Utilis pour faire clignoter l'alignement
    MainMenu: TMainMenu;
    MenuItemPartie: TMenuItem;
    MenuItemEdition: TMenuItem;
    MenuItemPartieNouvelle: TMenuItem;
    MenuItemPartieEnregistrer: TMenuItem;
    MenuItemPartieOuvrir: TMenuItem;
    MenuItemPartieQuitter: TMenuItem;
    MenuItemAPropos: TMenuItem;
    MenuItemDefaire: TMenuItem;
    MenuItemCouleurs: TMenuItem;
    MenuItemPartieEnregistrerSous: TMenuItem;
    PopupMenuJoue: TPopupMenu; // Menu droulant pour les commandes au clavier
    MenuItemColonne1: TMenuItem;
    MenuItemColonne2: TMenuItem;
    MenuItemColonne3: TMenuItem;
    MenuItemColonne4: TMenuItem;
    MenuItemColonne5: TMenuItem;
    MenuItemColonne6: TMenuItem;
    MenuItemColonne7: TMenuItem;
    StatusBar: TStatusBar;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    MenuItemRefaire: TMenuItem;
    MenuItemSeparator1: TMenuItem;
    procedure MenuItemRefaireClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FormResize(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure MenuItemPartieEnregistrerSousClick(Sender: TObject);
    procedure MenuItemPartieEnregistrerClick(Sender: TObject);
    procedure MenuItemPartieOuvrirClick(Sender: TObject);
    procedure MenuItemAProposClick(Sender: TObject);
    procedure MenuItemDefaireClick(Sender: TObject);
    procedure MenuItemColonneClick(Sender: TObject);
    procedure MenuItemPartieNouvelleClick(Sender: TObject);
    procedure MenuItemPartieQuitterClick(Sender: TObject);
    procedure MenuItemCouleursClick(Sender: TObject);
    procedure TimerAlignementTimer(Sender: TObject);
    procedure TimerChuteTimer(Sender: TObject);
  private
    // Calculs des dimensions en fonction de la taille de la fentre principale
    CoteCase, // Dimension d'une case
    HauteurGrille, LargeurGrille, // Dimensions de la grille
    x0, y0, // Coin Haut Gauche
    xChute, yChute, // coordonnes du pion en chute libre
    ColonneSouris : Integer; // Colonne sous la souris en temps rel
    FNomFichierCourant : String; // Nom de fichier en cours
    procedure AffecteNomFichier(const stNomFichier : String);
    property NomFichierCourant : String read FNomFichierCourant write AffecteNomFichier;
    function ChuteEnCours : Boolean;
    procedure ControleRegistre;
    procedure InitialisePartie;
    procedure CalculeDimensionsGrille;
    procedure DessineCase(const x : TColonne; const y : TLigne);
    procedure AfficheNumerosColonnes;
    procedure EvaluePossibilitesPlacement;
    procedure LachePion(const Colonne : TColonne);
    procedure MontreAlignement;
  public
    procedure RedessinePlateau;
  end;

var
  FormMain: TFormMain;
  Couleur : array [TCase] of TColor = (clBlue, clYellow, clRed);

implementation

{$R *.dfm}

Uses
  Registry, Math, couleurs_f;

const
  stTitrePrincipal : String = 'Fouquet''s Puissance 4';
  stVide           : String = '';
  stCleExtFp4      : String = '.fp4';
  stCleFp4         : String = 'Fp4';

var
  Jeu : TPuissance4;
  RapportJetonCase : Double = 0.9;

// ----------------------------------------------------------------------------
// Mthodes personnalises non "publies"
// ----------------------------------------------------------------------------
function TFormMain.ChuteEnCours : Boolean; // Indique si l'animation de chute est en cours.
begin
Result:=yChute<>0;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MontreAlignement; // Dclenche le clignotement des pions aligns
begin
MenuItemDefaire.Enabled:=False; // On interdit de dfaire
TimerAlignement.Enabled:=True; // On lance le clignotement pour mettre en vidence l'alignement des 4 pions
end;
// ----------------------------------------------------------------------------
procedure TFormMain.ControleRegistre;
var Reg : TRegistry;
begin // On contrle la base de registre sur le lien avec les fichiers de parties
Reg:=TRegistry.Create;
try
  try
    Reg.RootKey:=HKEY_CLASSES_ROOT;
    if not Reg.KeyExists(stCleExtFp4) then
      begin
      Reg.OpenKey(stCleExtFp4, true);
      Reg.WriteString(stVide, stCleFp4);
      Reg.CloseKey();
      end;
    Reg.OpenKey(stCleFp4, true);
    Reg.WriteString(stVide, 'Partie de Puissance 4');
    Reg.OpenKey('DefaultIcon', true);
    Reg.WriteString(stVide, Application.ExeName+',0');
    Reg.OpenKey(stCleFp4+'\shell', true);
    Reg.OpenKey('open', true);
    Reg.OpenKey('command', true);
    Reg.WriteString(stVide, Application.ExeName+' "%1"');
    Reg.CloseKey();
    Reg.CloseKey();
    Reg.CloseKey();
    Reg.CloseKey();
    Reg.CloseKey();
  except
    Application.MessageBox('Impossible d''ouvrir et/ou de mettre  jour la base de registres', pChar(stTitrePrincipal), MB_ICONHAND);
  end{try};
finally
  FreeAndNil(Reg);
end{try};
end;
// ----------------------------------------------------------------------------
procedure TFormMain.CalculeDimensionsGrille;
begin // Calculs des dimensions en fonction de la taille de la fentre principale
CoteCase:=Min(ClientWidth,ClientHeight) div 9;
HauteurGrille:=NbCasesParColonne*CoteCase;
LargeurGrille:=NbCasesParLigne*CoteCase;
x0:=(ClientWidth-LargeurGrille) div 2;
y0:=(ClientHeight-HauteurGrille) div 2;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.DessineCase(const x : TColonne; const y : TLigne);
begin
if Jeu=nil then
  Exit
else
  with Canvas,Jeu do
    begin
    // Case de la grille
    Brush.Color:=Couleur[cPersonne];
    FillRect(Rect(x0+(x-1)*CoteCase+1,
                  y0+(0+High(TLigne)-y)*CoteCase+1,
                  x0+(x-0)*CoteCase-1,
                  y0+(1+High(TLigne)-y)*CoteCase-1));
    Brush.Color:=IfThen(p[Tour,x,y]>cPersonne, Couleur[p[Tour,x,y]], self.Color);
    Pen.Color:=Self.Color;
    // Pion ou vide
    Ellipse(Rect(x0+Round((x-RapportJetonCase)*CoteCase)+1,
                 y0+Round((1+High(TLigne)-RapportJetonCase-y)*CoteCase)+1,
                 x0+Round((x-(1-RapportJetonCase))*CoteCase)-1,
                 y0+Round((0+High(TLigne)+RapportJetonCase-y)*CoteCase)-1));
    end{with}
end{DessineCase};
// ----------------------------------------------------------------------------
procedure TFormMain.AfficheNumerosColonnes; 
var x : TColonne;
    stNumero : String;
begin
  with Canvas do
    for x:=Low(TColonne) to High(TColonne) do
      begin
      stNumero:=IntToStr(x);
      Brush.Color:=self.Color;
      with Jeu do // On affiche avec les couleurs du prochain joueur ou du vainqueur
        Font.Color:=Couleur[TCase(IfThen(Vainqueur=cPersonne,
                                         1+Jeu.Tour mod 2,
                                         Integer(Vainqueur)))];
      TextOut(x0+(x-1)*CoteCase+(CoteCase-TextWidth(stNumero)) div 2,
              y0-TextHeight(stNumero),
              stNumero);
      end{for}
end;
// ----------------------------------------------------------------------------
procedure TFormMain.RedessinePlateau;
var
  i : Integer;
  x : TColonne;     // Abscisse (colonne) de la case  dessiner
  y : TLigne;       // Ordonne (ligne) de la case  dessiner
  begin
with Canvas do
  begin
  Pen.Color:=Couleur[cPersonne];
  Brush.Color:=Color;
  FillRect(Rect(0,0,ClientWidth,ClientHeight));
  if Jeu=nil then Exit; // On affiche rien si pas de jeu en cours
  // 0. Les numros de colonne
  AfficheNumerosColonnes;
  // 1. La grille
  for i:=0 to Max(High(TLigne),High(TColonne)) do
    begin
    MoveTo(x0+i*CoteCase, y0);
    LineTo(x0+i*CoteCase, y0+HauteurGrille);
    if i<Max(High(TLigne),High(TColonne)) then
      begin
      Brush.Color:=clWhite;
      MoveTo(x0, y0+i*CoteCase);
      LineTo(x0+LargeurGrille, y0+i*CoteCase);
      end
    end{for};
  // 2. Les pions
  with Jeu do
    for x:=Low(TColonne) to High(TColonne) do
      for y:=Low(TLigne) to High(TLigne) do
        DessineCase(x, y);
  end{with Canvas}
end{RedessinePlateau};
// ----------------------------------------------------------------------------
procedure TFormMain.EvaluePossibilitesPlacement;
var x : TColonne;
begin // Permet d'empcher le choix d'une colonne pleine
// On force une rvaluation au niveau de la souris au prochain dplacement du pointeur.
for x:=Low(TColonne) to High(TColonne) do
  (FindComponent('MenuItemColonne'+IntToStr(x)) as TMenuItem).Enabled:=
    (Jeu<>nil) and
    (Jeu.Vainqueur=cPersonne) and
    (Jeu.p[Jeu.Tour,x,High(TLigne)]=cPersonne);
if (ColonneSouris>=Low(Tcolonne)) and
   (ColonneSouris<=High(TColonne)) and
   not (FindComponent('MenuItemColonne'+IntToStr(ColonneSouris)) as TMenuItem).Enabled then
  begin
  ColonneSouris:=0;
  Cursor:=crNo;
  end
end{EvaluePossibilitesPlacement};
// ----------------------------------------------------------------------------
procedure TFormMain.LachePion(const Colonne : TColonne);
begin  // Lance l'animation de chute du pion
xChute:=Colonne;
yChute:=High(TLigne);
Jeu.PrepareTourSuivant;
MenuItemPartieEnregistrer.Enabled:=True;
MenuItemPartieEnregistrerSous.Enabled:=True;
TimerChute.Enabled:=True;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.AffecteNomFichier(const stNomFichier : String);
begin
FNomFichierCourant:=stNomFichier;
if FNomFichierCourant='' then
  Caption:=stTitrePrincipal
else
  Caption:=Format('%s [%s]', [stTitrePrincipal, FNomFichierCourant]); // Tout l'intrt de la property NomFichierCourant
end;
// ----------------------------------------------------------------------------
procedure TFormMain.InitialisePartie;
begin // Initialisation avant dmarrage d'une partie
with TimerAlignement do
  begin
  Enabled:=False;
  Tag:=1
  end;
MenuItemDefaire.Enabled:=False;
if Jeu<>nil then FreeAndNil(Jeu);
xChute:=0;
yChute:=0;
ColonneSouris:=0;
NomFichierCourant:='';
end;
// ----------------------------------------------------------------------------
// Evnements
// ----------------------------------------------------------------------------
procedure TFormMain.FormCreate(Sender: TObject);
begin
ControleRegistre;
Jeu:=nil;
InitialisePartie;
EvaluePossibilitesPlacement;
Caption:=stTitrePrincipal;
if ParamCount>0 then // Paramtre facultatif : nom d'une partie sauvegarde
  if FileExists(ParamStr(1)) then
    begin
    OpenDialog.Tag:=1; // On ne demande pas le fichier (ouverture automatique)
    try
      MenuItemPartieOuvrirClick(MenuItemPartieOuvrir);
    finally
      OpenDialog.Tag:=0; // On remet la demande de fichier
    end{try}
    end{if FileExists}
  else
    Application.MessageBox('Le paramtre pass en ligne de commande n''est pas un nom de fichier correct ou existant.', 'Impossible d''ouvrir le fichier automatiquement', MB_ICONHAND);
end;
// ----------------------------------------------------------------------------
procedure TFormMain.FormDestroy(Sender: TObject);
begin
InitialisePartie;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.FormPaint(Sender: TObject);
begin
RedessinePlateau;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.FormResize(Sender: TObject);
begin
 CalculeDimensionsGrille;
 RedessinePlateau;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
if (X<x0) or (X>=x0+High(TColonne)*CoteCase) or
   (Y<y0) or (Y>=y0+High(TLigne)*CoteCase) then
  begin // Le pointeur N'est PAS sur la grille...
  if ColonneSouris<>0 then
    begin
    ColonneSouris:=0;
    if (Jeu=Nil) or (Jeu.Vainqueur=cPersonne) then
      StatusBar.SimpleText:='';
    Cursor:=crNo;
    end
  end
else // Le pointeur est sur la grille... On calcule la colonne
  if ColonneSouris<>1+(X-x0) div CoteCase then
    begin
    ColonneSouris:=1+(X-x0) div CoteCase;
    if not (FindComponent('MenuItemColonne'+IntToStr(ColonneSouris)) as TMenuItem).Enabled then
      begin
      ColonneSouris:=0;
      if (Jeu=Nil) or (Jeu.Vainqueur=cPersonne) then
        StatusBar.SimpleText:='';
      Cursor:=crNo;
      end
    else
      begin
      StatusBar.SimpleText:=Format('Tapez %d ou cliquez pour lcher un pion dans la colonne %d.', [ColonneSouris, ColonneSouris]);
      Cursor:=crDefault;
      end
    end
end;
// ----------------------------------------------------------------------------
procedure TFormMain.FormClick(Sender: TObject);
var MenuItem : TMenuItem;
begin
 if (ColonneSouris>=Low(TColonne)) and
    (ColonneSouris<=High(TColonne)) then
   begin // La souris est sur une des 7 colonnes de la grille...
   MenuItem:=FindComponent('MenuItemColonne'+IntToStr(ColonneSouris)) as TMenuItem;
   if MenuItem.Enabled then // La colonne a encore une place de libre
     MenuItemColonneClick(MenuItem); // On lche le jeton
   end;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin // On demande confirmation si une partie est commence.
CanClose:=(Jeu=nil) or
          (Jeu.Tour=0) or
          (Application.MessageBox('tes-vous sr(e) de vouloir abandonner la partie en cours ?', 'Confirmation', MB_ICONQUESTION+MB_YESNO)=IDYES)
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemPartieNouvelleClick(Sender: TObject);
begin
InitialisePartie;
Jeu:=TPuissance4.Create;
try
  CalculeDimensionsGrille;
  RedessinePlateau;
  EvaluePossibilitesPlacement;
except
  Application.MessageBox('Impossible de commencer une nouvelle partie (mmoire insuffisante ?)', 'Problme de mmoire', MB_ICONHAND);
end{try};
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemColonneClick(Sender: TObject);
var x : TColonne;
begin
if ChuteEnCours then Exit; // On attend que le pion arrive  destination avant le prochain coup.
x:=(Sender as TMenuItem).Tag;
LachePion(x);
if not MenuItemDefaire.Enabled then
  MenuItemDefaire.Enabled:=True;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemCouleursClick(Sender: TObject);
begin
if FormCouleurs.ShowModal=mrOk then
  RedessinePlateau;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemPartieQuitterClick(Sender: TObject);
begin
Close;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemDefaireClick(Sender: TObject);
begin
if Jeu.Tour>0 then
  begin
  Dec(Jeu.Tour);
  MenuItemRefaire.Enabled:=True;
  RedessinePlateau
  end;
MenuItemDefaire.Enabled:=Jeu.Tour>0;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemRefaireClick(Sender: TObject);
begin
if Jeu.Tour<=Jeu.DernierTour then
  begin
  Inc(Jeu.Tour);
  MenuItemDefaire.Enabled:=True;
  RedessinePlateau
  end;
MenuItemRefaire.Enabled:=Jeu.Tour<Jeu.DernierTour;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemAProposClick(Sender: TObject);
begin
Application.MessageBox('Fouquet''s Puissance 4'#13#13'Licence Publique Gnrale 2'#13#13'Dvelopp avec Delphi 2005'#13'par Patrice Fouquet'#13#13'patquoi@free.fr'#13'http://patquoi.free.fr',
                       ' propos de FP4...',
                       MB_ICONINFORMATION)
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemPartieOuvrirClick(Sender: TObject);
begin
if (OpenDialog.Tag=1) or // Ouverture automatique en paramtre de ligne de commande
   OpenDialog.Execute then
  begin
  InitialisePartie;
  if OpenDialog.Tag=0 then // Si le fichier a t slectionn (ouverture non automatique)
    NomFichierCourant:=OpenDialog.FileName
  else
    NomFichierCourant:=ParamStr(1);
  Jeu:=TPuissance4.Create;
  try
    if not Jeu.OuvrePartie(NomFichierCourant) then
      begin
      Application.MessageBox('Impossible d''ouvrir la partie sauvegarde !', 'Erreur  l''ouverture', MB_ICONHAND);
      InitialisePartie;
      end;
    EvaluePossibilitesPlacement;
    RedessinePlateau;
  except
    Application.MessageBox('Impossible de dmarrer la partie ouverte (mmoire insuffisante ?)', 'Problme de mmoire', MB_ICONHAND);
  end{try};
  end{if}
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemPartieEnregistrerClick(Sender: TObject);
begin
if (NomFichierCourant<>'') or SaveDialog.Execute then
  begin
  if NomFichierCourant='' then
    NomFichierCourant:=SaveDialog.FileName;
  if not Jeu.EnregistrePartie(NomFichierCourant) then
    Application.MessageBox('Impossible d''enregistrer la partie !', 'Erreur  l''enregistrement', MB_ICONHAND);
  end;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.MenuItemPartieEnregistrerSousClick(Sender: TObject);
begin
if SaveDialog.Execute then
  begin
  NomFichierCourant:=SaveDialog.FileName;
  MenuItemPartieEnregistrerClick(MenuItemPartieEnregistrerSous);
  end;
end;
// ----------------------------------------------------------------------------
procedure TFormMain.TimerChuteTimer(Sender: TObject);
begin
with Jeu do
  begin
  p[Tour, xChute, yChute]:=JoueurCourant;
  DessineCase(xChute, yChute);
  if (yChute<High(TLigne)) then
    begin
    p[Tour, xChute, yChute+1]:=cPersonne;
    DessineCase(xChute, yChute+1);
    end;
  if (yChute=Low(TLigne)) or
     (p[Tour, xChute, yChute-1]>cPersonne) then
    begin
    TimerChute.Enabled:=False;
    if QuatrePionsAlignes(xChute, yChute)=dIndefinie then
      begin
      AfficheNumerosColonnes;
      EvaluePossibilitesPlacement;
      xChute:=0; yChute:=0;
      end{if Quatre}
    else
      begin
      Vainqueur:=JoueurCourant;
      StatusBar.SimpleText:=Format('C''est gagn ! Choisir Partie | Nouvelle (%s) pour rejouer.',
                                   [ShortCutToText(MenuItemPartieNouvelle.ShortCut)]);
      Cursor:=crNo;
      ColonneSouris:=0;
      EvaluePossibilitesPlacement;
      MontreAlignement;
      end;
    end
  else
    Dec(yChute);
  end{with Jeu}
end;
// ----------------------------------------------------------------------------
procedure TFormMain.TimerAlignementTimer(Sender: TObject);
var i : Integer;
begin
TimerAlignement.Tag:=1-TimerAlignement.Tag;
with Jeu do
  for i:=-NbMemesPionsAvant to NbMemesPionsApres do
    begin
    p[Tour, xChute+i*dx[dAlignement],
            yChute+i*dy[dAlignement]]:=TCase(TimerAlignement.Tag*Integer(Vainqueur));
    DessineCase(xChute+i*dx[dAlignement],
                yChute+i*dy[dAlignement]);
    end{for}
end;
// ----------------------------------------------------------------------------
end.
